# libraries
import numpy as np
import pandas as pd
from datetime import datetime, timedelta

from tank_forecaster.decomp import generic_daily_decomp


def format_tank(dict_input, freq: str = "30min"):

    if dict_input == [] or dict_input is None:
        return None

    else:
        dict_input = sorted(dict_input, key=lambda x: x["read_time"])
        dict_input = [d for d in dict_input if d["volume"]]
        df = pd.DataFrame(dict_input)
        df.read_time = pd.to_datetime(df.read_time)
        df["y"] = df.volume.diff() * -1
        df = df[["read_time", "y"]].rename(columns={"read_time": "ds"})
        df = df[df.y < 800].copy()
        df = df.set_index("ds").groupby(pd.Grouper(freq=freq)).sum()
        df = df[df.y > 0].copy()

        if len(df) == 2:
            return None

        df.reset_index(level=0, inplace=True)
        output = df.iloc[1:-1]

        return output



def format_sales(dict_input):

    if dict_input == [] or dict_input is None:
        return None

    else:
        dict_input = sorted(dict_input, key=lambda x: x['date'])
        dat = pd.DataFrame(dict_input)

        # only same product type
        if 'tank_type' in dat:
            df = dat[dat['tank_type'] == dat.tank_type.iloc[-1]]
        else:
            df = dat

        df = df[['date', 'sales']]
        df.sales = df.sales.clip(lower=df.sales.quantile(0.015), upper=df.sales.quantile(0.995))
        df.rename(columns={'date': 'ds', 'sales': 'y'}, inplace=True)
        df['ds'] = pd.to_datetime(df['ds'], format='%Y-%m-%d')

    return df


def gen_past(df, freq='30min', periods=432):

    if len(df) == 0 or df is None:
        now = datetime.now()
        end_0 = now - (now - datetime.min) % timedelta(minutes=30)

    else:
        end_0 = df.iloc[-1, 0]

    past = pd.date_range(end=end_0, freq=freq, periods=periods)
    past = pd.DataFrame(past)
    past.rename(columns={0: 'ds'}, inplace=True)
    past['ds'] = pd.to_datetime(past['ds'], format='%Y-%m-%d')

    return past


def validate_tank(tank_data, sales_data):

    if len(tank_data) == 0 or tank_data is None:
        return None

    # add day of year for tank data / sales data
    sales_data['doy'] = sales_data['ds'].dt.dayofyear
    sales_data = sales_data.iloc[-100:]

    # create validation dataframe
    val_matrix = np.zeros((432, 2))
    val_df = pd.DataFrame(val_matrix,
                          columns=['ds',
                                   'doy'])

    # dynamically generate timestamps
    val_df['ds'] = gen_past(tank_data, freq='30min', periods=432)
    val_df['doy'] = val_df.ds.dt.dayofyear
    val_df['ts'] = val_df.ds.dt.time

    val_df = pd.merge(left=val_df, right=generic_daily_decomp, how='left', left_on='ts', right_on='ts')
    val_df = pd.merge(left=val_df, right=tank_data, how='left', left_on='ds', right_on='ds')
    val_df = pd.merge(left=val_df, right=sales_data, how='left', left_on='doy', right_on='doy')

    val_df.rename(columns={'ds_x': 'ds', 'y_x': 'actual', 'y_y': 'daily'}, inplace=True)
    val_df = val_df.loc[:, ['ds', 'ts', 'actual', 'daily', 'daily_multi']]

    # calculate generalized daily tank sticking curve from sales data
    val_df['base_est'] = (val_df.loc[:, 'daily'] * (1 / 48))
    est = val_df['base_est'] * val_df['daily_multi']
    val_df['estimated'] = est

    val_df = val_df.loc[:, ['ds', 'actual', 'estimated']]

    # data abnormalities adjustment
    diff = val_df.estimated - val_df.actual
    threshold = diff.abs().quantile(0.95)

    mask_adj = diff.abs() > threshold
    mask_tank = [not i for i in mask_adj]

    # val_df.loc[mask_adj, 'output'] = val_df.loc[mask_adj, 'adj_sales']
    # val_df.loc[mask_tank, 'output'] = val_df.loc[mask_tank, 'tank_vals']

    # missing data imputation
    mask_nan = val_df.actual.isna()
    mask_not_nan = [not i for i in mask_nan]

    val_df.loc[mask_nan, 'y'] = val_df.loc[mask_nan, 'estimated']
    val_df.loc[mask_not_nan, 'y'] = val_df.loc[mask_not_nan, 'actual']

    # for testing purposes, uncomment this
    # return val_df

    # for use in forecasting, uncomment this
    val_df.rename(columns={'ds_x': 'ds'}, inplace=True)

    return val_df[['ds', 'y']]



import pandas as pd
import numpy as np

from tank_forecaster import forecaster


def test_far_none():
    x = forecaster.forecast_far(None, daily_lift_est=1000)
    assert type(x) is pd.core.frame.DataFrame
    assert 'ds' in x and 'yhat' in x
    assert len(x) == 90
    assert type(x.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
    assert type(x.yhat.iloc[-1]) is np.float64


def test_far_little_data(sales_little_data):
    x = forecaster.forecast_far(sales_little_data)
    assert type(x) is pd.core.frame.DataFrame
    assert 'ds' in x and 'yhat' in x
    assert len(x) == 90
    assert type(x.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
    assert type(x.yhat.iloc[-1]) is np.float64


def test_far_proper():
    pass


def test_near_none():
    x = forecaster.forecast_near(None, daily_lift_est=1000)
    assert type(x) is pd.core.frame.DataFrame
    assert len(x) == 144
    assert 'ds' in x and 'yhat' in x
    assert type(x.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
    assert type(x.yhat.iloc[-1]) is np.float64


def test_near_little_data(tank_little_data):
    x = forecaster.forecast_near(tank_little_data)
    assert type(x) is pd.core.frame.DataFrame
    assert len(x) == 144
    assert 'ds' in x and 'yhat' in x
    assert type(x.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
    assert type(x.yhat.iloc[-1]) is np.float64


def test_near_proper():
    pass



import pandas as pd
import numpy as np
from fbprophet import Prophet

# generic yearly / weekly / daily decompositions
generic_weekly_decomp = pd.Series([0.95, 1, 1.05, 1.05, 1.20, 0.95, 0.8])

generic_yearly_decomp = pd.Series([0.95, 0.95, *np.repeat(0.9, 10), 0.925, 0.95, 0.975,
                                   *np.repeat(1, 7), 1.025, 1.05, *np.repeat(1.075, 11),
                                   1.05, 1.025, *np.repeat(1, 7), *np.repeat(0.95, 9)])
len(generic_weekly_decomp)
len(generic_yearly_decomp)

generic_daily_decomp = pd.DataFrame(columns=['ts', 'vals'])
generic_daily_decomp['ds'] = pd.date_range(start='00:00:00', end='23:30:00', freq='30min')
generic_daily_decomp['ts'] = generic_daily_decomp.ds.dt.time
generic_daily_decomp['daily_multi'] = pd.Series([1, 0.85, 0.7, 0.55, 0.4,
                                  0.25, 0.1, *np.repeat(0, 9),
                                  0.1, 0.2, 0.3, 0.4, 0.5,
                                  0.6, 0.7, 0.8, 0.9, 1.0,
                                  1.1, 1.2, 1.3, 1.4, 1.5,
                                  1.6, 1.7, 1.8, 1.9, *np.repeat(2, 4),
                                  1.9, 1.8, 1.7, 1.6, 1.5,
                                  1.4, 1.3, 1.2, 1.1])
generic_daily_decomp = generic_daily_decomp.loc[:, ['ts', 'daily_multi']]


def decompose_sales(df):

    if df is None:
        return generic_yearly_decomp, generic_weekly_decomp

    if len(df) == 0:
        return generic_yearly_decomp, generic_weekly_decomp

    else:
        m = Prophet(changepoint_prior_scale=0.05,
                    daily_seasonality=False,
                    weekly_seasonality=True,
                    yearly_seasonality=True,
                    seasonality_mode='multiplicative')

        m.fit(df)
        future = m.make_future_dataframe(periods=1, freq='1D')
        forecast = m.predict(future)
        forecast_reduced = forecast.loc[:, ['ds', 'yearly', 'weekly']]
        forecast_reduced['dow'] = forecast_reduced.ds.dt.weekday
        forecast_reduced['doy'] = forecast_reduced.ds.dt.dayofyear
        forecast_reduced['woy'] = forecast_reduced.ds.dt.week  # deprecated?
        week_trend = forecast_reduced.groupby('dow')['weekly'].mean() + 1
        forecast_reduced = forecast_reduced.iloc[-366:]
        forecast_reduced.sort_values(by=['woy', 'dow'], inplace=True)
        year_trend = forecast_reduced.groupby('woy').mean()
        year_trend = year_trend['yearly'] + 1

    return year_trend, week_trend


# tank history decomposition for daily trends
def decompose_tank(tank_history):

    m = Prophet(changepoint_prior_scale=0.05,
                daily_seasonality=True,
                weekly_seasonality=True,
                yearly_seasonality=False,
                seasonality_mode='multiplicative')

    mod = m.fit(tank_history)
    future = m.make_future_dataframe(periods=1, freq='30min')
    forecast = m.predict(future)
    forecast_reduced = forecast.loc[:, ['ds', 'daily', 'daily_lower', 'daily_upper']]
    forecast_reduced['time'] = forecast_reduced['ds'].dt.time
    forecast_reduced.drop(columns='ds', inplace=True)
    day_trend = forecast_reduced.groupby('time').mean()
    day_trend.sort_values(by=['time'], inplace=True)

    for field in ["daily", "daily_lower", "daily_upper"]:
        day_trend.loc[:, field] = day_trend.loc[:, field] + 1

    return day_trend


# holiday decomposition - needs improvement for consistent date mapping
def decompose_holiday(input):
    if (type(input) != pd.core.frame.DataFrame):
        return print('error - input is not a DataFrame')
    else:
        event_1 = pd.DataFrame({
            'holiday': 'event_1_name',
            'ds': pd.to_datetime(['2020-03-01']),
            'lower_window': 0,
            'upper_window': 0,
        })
        event_2 = pd.DataFrame({
            'holiday': 'event_2_name',
            'ds': pd.to_datetime(['2020-03-02']),
            'lower_window': 0,
            'upper_window': 0,
        })
        holidays = pd.concat([event_1, event_2])
        m = Prophet(holidays=holidays,
                    changepoint_prior_scale=0.05,
                    daily_seasonality=False,
                    weekly_seasonality=True,
                    yearly_seasonality=True,
                    seasonality_mode='multiplicative')
        m.add_country_holidays(country_name='US')
        m.fit(df)
        future = m.make_future_dataframe(periods=1, freq='1D')
        forecast = m.predict(future)
        forecast_reduced = forecast.loc[:, ['ds', 'holidays']]

    return forecast_reduced




#!/usr/bin/env python

import numpy as np
import pandas as pd

from tank_forecaster import validation


def test_working():
    x = True
    assert x


def test_format_tank_none():
    x = validation.format_tank(None)
    assert x is None


def test_format_tank_empty():
    x = validation.format_tank([])
    assert x is None


def test_format_tank_two_entries(tank_two_entry_input):
    x = validation.format_tank(tank_two_entry_input)
    assert x is None


def test_format_tank_proper(tank_little_input):
    x = validation.format_tank(tank_little_input)
    assert type(x) is pd.core.frame.DataFrame
    assert len(x['ds']) != 0
    assert len(x['y']) != 0
    assert type(x.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
    assert type(x.y.iloc[-1]) is np.float64


def test_format_sales_none():
    x = validation.format_sales(None)
    assert x is None


def test_format_sales_empty():
    x = validation.format_sales([])
    assert x is None


def test_format_sales_proper(sales_little_input):
    x = validation.format_sales(sales_little_input)
    assert type(x) is pd.core.frame.DataFrame
    assert len(x['ds']) != 0
    assert len(x['y']) != 0
    assert type(x.ds.iloc[-1]) is pd._libs.tslibs.timestamps.Timestamp
    assert type(x.y.iloc[-1]) is np.float64
    # change
